"use strict";
/**
 * @dev Class for managing secp256k1 keys and performing operations with them
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.KeyPair = void 0;
const secp256k1_1 = require("@noble/secp256k1");
const ethers_1 = require("../ethers");
const RandomNumber_1 = require("./RandomNumber");
const utils_1 = require("../utils/utils");
// List of private or public keys that we disallow initializing a KeyPair instance with, since they will lead to
// unrecoverable funds.
const blockedKeys = [
    '0x0000000000000000000000000000000000000000000000000000000000000000',
    '0x0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000', // public key of all zeroes
];
/**
 * @notice Private helper method to return the shared secret for a given private key and public key
 * @param privateKey Private key as hex string with 0x prefix
 * @param publicKey Uncompressed public key as hex string with 0x04 prefix
 * @returns 32-byte shared secret as 66 character hex string
 */
function getSharedSecret(privateKey, publicKey) {
    if (privateKey.length !== utils_1.lengths.privateKey || !(0, ethers_1.isHexString)(privateKey))
        throw new Error('Invalid private key');
    if (publicKey.length !== utils_1.lengths.publicKey || !(0, ethers_1.isHexString)(publicKey))
        throw new Error('Invalid public key');
    (0, utils_1.assertValidPoint)(publicKey);
    (0, utils_1.assertValidPrivateKey)(privateKey);
    // We use sharedSecret.slice(2) to ensure the shared secret is not dependent on the prefix, which enables
    // us to uncompress ephemeralPublicKey from Umbra.sol logs as explained in comments of getUncompressedFromX.
    // Note that a shared secret is really just a point on the curve, so it's an uncompressed public key
    const sharedSecret = (0, secp256k1_1.getSharedSecret)(privateKey.slice(2), publicKey.slice(2), true);
    const sharedSecretHex = secp256k1_1.utils.bytesToHex(sharedSecret); // Has 04 prefix but not 0x.
    return (0, ethers_1.sha256)(`0x${sharedSecretHex.slice(2)}`); // TODO Update to use noble-hashes?
}
class KeyPair {
    /**
     * @notice Creates new instance from a public key or private key
     * @param key Can be either (1) hex public key with 0x04 prefix, or (2) hex private key with 0x prefix
     */
    constructor(key) {
        this.privateKeyHex = null; // Private key as hex string with 0x prefix, or null if not provided
        if (typeof key !== 'string' || !(0, ethers_1.isHexString)(key)) {
            throw new Error('Key must be a string in hex format with 0x prefix');
        }
        if (blockedKeys.includes(key)) {
            throw new Error('Cannot initialize KeyPair with the provided key');
        }
        if (key.length === utils_1.lengths.privateKey) {
            // Private key provided
            (0, utils_1.assertValidPrivateKey)(key);
            this.privateKeyHex = key;
            const publicKey = (0, secp256k1_1.getPublicKey)(this.privateKeyHexSlim);
            this.publicKeyHex = `0x${secp256k1_1.utils.bytesToHex(publicKey)}`; // Has 0x04 prefix, other forms computed as getters.
        }
        else if (key.length === utils_1.lengths.publicKey) {
            // Public key provided
            (0, utils_1.assertValidPoint)(key); // throw if point is not on curve
            this.publicKeyHex = key; // Save off public key, other forms computed as getters
        }
        else {
            throw new Error('Key must be a 66 character hex private key or a 132 character hex public key');
        }
    }
    // ===================================================== GETTERS =====================================================
    /**
     * @notice Returns the private key as a hex string without the 0x prefix
     */
    get privateKeyHexSlim() {
        return this.privateKeyHex ? this.privateKeyHex.slice(2) : null;
    }
    /**
     * @notice Returns the uncompressed public key as a hex string without the 0x prefix
     */
    get publicKeyHexSlim() {
        return this.publicKeyHex.slice(2);
    }
    /**
     * @notice Returns checksum address derived from this key
     */
    get address() {
        return (0, ethers_1.computeAddress)(this.publicKeyHex);
    }
    // ============================================= ENCRYPTION / DECRYPTION =============================================
    /**
     * @notice Encrypt a number with the instance's public key
     * @param randomNumber Number as instance of RandomNumber class
     * @returns Hex strings of uncompressed 65 byte public key and 32 byte ciphertext
     */
    encrypt(number) {
        if (!(number instanceof RandomNumber_1.RandomNumber)) {
            throw new Error('Must provide instance of RandomNumber');
        }
        // Get shared secret to use as encryption key
        const ephemeralPrivateKey = secp256k1_1.utils.randomPrivateKey();
        const ephemeralPublicKey = secp256k1_1.Point.fromPrivateKey(ephemeralPrivateKey);
        const ephemeralPrivateKeyHex = `0x${secp256k1_1.utils.bytesToHex(ephemeralPrivateKey)}`;
        const ephemeralPublicKeyHex = `0x${ephemeralPublicKey.toHex()}`;
        const sharedSecret = getSharedSecret(ephemeralPrivateKeyHex, this.publicKeyHex);
        // XOR random number with shared secret to get encrypted value
        const ciphertextBN = number.value.xor(sharedSecret);
        const ciphertext = (0, ethers_1.hexZeroPad)(ciphertextBN.toHexString(), 32); // 32 byte hex string with 0x prefix
        return { ephemeralPublicKey: ephemeralPublicKeyHex, ciphertext };
    }
    /**
     * @notice Decrypt a random number with the instance's private key and return the plaintext
     * @param output Output from the encrypt method, which can be constructed from on-chain events
     * @returns Decrypted ciphertext as hex string
     */
    decrypt(output) {
        const { ephemeralPublicKey, ciphertext } = output;
        if (!ephemeralPublicKey || !ciphertext) {
            throw new Error('Input must be of type EncryptedPayload to decrypt');
        }
        if (!this.privateKeyHex) {
            throw new Error('KeyPair has no associated private key to decrypt with');
        }
        (0, utils_1.assertValidPoint)(ephemeralPublicKey); // throw if point is not on curve
        // Get shared secret to use as decryption key, then decrypt with XOR
        const sharedSecret = getSharedSecret(this.privateKeyHex, ephemeralPublicKey);
        const plaintext = ethers_1.BigNumber.from(ciphertext).xor(sharedSecret);
        return (0, ethers_1.hexZeroPad)(plaintext.toHexString(), 32);
    }
    // =============================================== ELLIPTIC CURVE MATH ===============================================
    /**
     * @notice Returns new KeyPair instance after multiplying this public key by some value
     * @param value number to multiply by, as RandomNumber or hex string with 0x prefix
     */
    mulPublicKey(value) {
        if (!(value instanceof RandomNumber_1.RandomNumber) && typeof value !== 'string') {
            throw new Error('Input must be instance of RandomNumber or string');
        }
        if (typeof value === 'string' && !value.startsWith('0x')) {
            throw new Error('Strings must be in hex form with 0x prefix');
        }
        // Parse number based on input type
        const number = (0, ethers_1.isHexString)(value)
            ? BigInt(value) // provided a valid hex string
            : BigInt(value.asHex); // provided RandomNumber
        // Perform the multiplication and return new KeyPair instance
        const publicKey = secp256k1_1.Point.fromHex(this.publicKeyHexSlim).multiply(number);
        return new KeyPair(`0x${publicKey.toHex()}`);
    }
    /**
     * @notice Returns new KeyPair instance after multiplying this private key by some value
     * @param value number to multiply by, as class RandomNumber or hex string with 0x prefix
     */
    mulPrivateKey(value) {
        if (!(value instanceof RandomNumber_1.RandomNumber) && typeof value !== 'string') {
            throw new Error('Input must be instance of RandomNumber or string');
        }
        if (typeof value === 'string' && !(0, ethers_1.isHexString)(value)) {
            throw new Error('Strings must be in hex form with 0x prefix');
        }
        if (!this.privateKeyHex) {
            throw new Error('KeyPair has no associated private key');
        }
        // Parse number based on input type
        const number = (0, ethers_1.isHexString)(value)
            ? BigInt(value) // provided a valid hex string
            : BigInt(value.asHex); // provided RandomNumber
        // Get new private key. Multiplication gives us an arbitrarily large number that is not necessarily in the domain
        // of the secp256k1 curve, so then we use modulus operation to get in the correct range.
        const privateKeyBigInt = (BigInt(this.privateKeyHex) * number) % secp256k1_1.CURVE.n;
        const privateKey = (0, ethers_1.hexZeroPad)(ethers_1.BigNumber.from(privateKeyBigInt).toHexString(), 32); // convert to 32 byte hex
        return new KeyPair(privateKey); // return new KeyPair instance
    }
    // ================================================= STATIC METHODS ==================================================
    /**
     * @notice Generate KeyPair instance asynchronously from a transaction hash
     * @param txHash Transaction hash to recover public key from
     * @param provider ethers provider to use
     */
    static async instanceFromTransaction(txHash, provider) {
        if (typeof txHash !== 'string' || txHash.length !== utils_1.lengths.txHash) {
            throw new Error('Invalid transaction hash provided');
        }
        const publicKeyHex = await (0, utils_1.recoverPublicKeyFromTransaction)(txHash, provider);
        return new KeyPair(publicKeyHex);
    }
    /**
     * @notice Takes an uncompressed public key and returns the compressed public key
     * @param publicKey Uncompressed public key, as hex string starting with 0x
     * @returns Object containing the prefix as an integer and compressed public key as hex, as separate parameters
     */
    static compressPublicKey(publicKey) {
        (0, utils_1.assertValidPoint)(publicKey);
        const compressedPublicKey = secp256k1_1.Point.fromHex(publicKey.slice(2)).toHex(true);
        return {
            prefix: Number(compressedPublicKey[1]),
            pubKeyXCoordinate: `0x${compressedPublicKey.slice(2)}`,
        };
    }
    /**
     * @notice Given the x-coordinate of a public key, without the identifying prefix bit, returns
     * the uncompressed public key assuming the identifying bit is 02
     * @dev We don't know if the identifying bit is 02 or 03 when uncompressing for the scanning use case, but it
     * doesn't actually matter since we are not deriving an address from the public key. We use the public key to
     * compute the shared secret to decrypt the random number, and since that involves multiplying this public key
     * by a private key, we can ensure the result is the same shared secret regardless of whether we assume the 02 or
     * 03 prefix by using the compressed form of the hex shared secret and ignoring the prefix. Therefore if no prefix
     * is provided, we can assume 02, and it's up to the user to make sure they are using this method safely. This is
     * done because it saves gas in the Umbra contract
     * @param pkx x-coordinate of compressed public key, as BigNumber or hex string
     * @param prefix Prefix bit, must be 2 or 3
     */
    static getUncompressedFromX(pkx, prefix = undefined) {
        // Converting `pkx` to a BigNumber will throw if the value cannot be safely converted to a BigNumber, i.e. if the
        // value is of type Number and larger than Number.MAX_SAFE_INTEGER.
        pkx = ethers_1.BigNumber.from(pkx);
        // pkx was validated, now we decompress it.
        const hexWithoutPrefix = (0, ethers_1.hexZeroPad)(ethers_1.BigNumber.from(pkx).toHexString(), 32).slice(2); // pkx as hex string without 0x prefix
        if (!prefix) {
            // Only safe to use this branch when uncompressed key is using for scanning your funds
            return `0x${secp256k1_1.Point.fromHex(`02${hexWithoutPrefix}`).toHex()}`;
        }
        const hexWithPrefix = `0${Number(prefix)}${hexWithoutPrefix}`;
        return `0x${secp256k1_1.Point.fromHex(hexWithPrefix).toHex()}`;
    }
}
exports.KeyPair = KeyPair;
//# sourceMappingURL=KeyPair.js.map