"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getPublicKeys = exports.namehash = exports.cnsKeyPathViewing = exports.cnsKeyPathSpending = void 0;
/**
 * @dev Functions for interacting with the Unstoppable Domains Crypto Name Service (CNS)
 */
const secp256k1_1 = require("@noble/secp256k1");
const resolution_1 = require("@unstoppabledomains/resolution");
const constants_1 = require("./constants");
const ethers_1 = require("../ethers");
exports.cnsKeyPathSpending = 'crypto.ETH.umbra.spending_public_key';
exports.cnsKeyPathViewing = 'crypto.ETH.umbra.viewing_public_key';
/**
 * @notice Computes CNS namehash of the input CNS domain, normalized to CNS compatibility
 * @param name CNS domain, e.g. myname.crypto
 * @param resolution Resolution instance of @unstoppabledomains/resolution
 */
function namehash(name, resolution) {
    if (typeof name !== 'string') {
        throw new Error('Name must be a string');
    }
    return resolution.namehash(name, resolution_1.NamingServiceName.UNS);
}
exports.namehash = namehash;
/**
 * @notice For a given CNS domain, returns the public keys, or throws if they don't exist
 * @param name CNS domain, e.g. myname.crypto
 * @param provider ethers provider to use
 * @param resolution Resolution instance of @unstoppabledomains/resolution
 */
async function getPublicKeys(name, provider, resolution) {
    // Read compressed public keys
    const domainNamehash = resolution.namehash(name, resolution_1.NamingServiceName.UNS);
    const resolverAddress = await resolution.resolver(name);
    const cnsResolver = new ethers_1.Contract(resolverAddress, constants_1.CNS_RESOLVER_ABI, provider);
    const [compressedSpendingPublicKey, compressedViewingPublicKey] = await cnsResolver.getMany([exports.cnsKeyPathSpending, exports.cnsKeyPathViewing], domainNamehash);
    if (!compressedSpendingPublicKey || !compressedViewingPublicKey) {
        throw new Error(`Public keys not found for ${name}. User must setup their Umbra account`);
    }
    // Return uncompressed public keys
    const spendingPublicKey = `0x${secp256k1_1.Point.fromHex(compressedSpendingPublicKey.slice(2)).toHex()}`;
    const viewingPublicKey = `0x${secp256k1_1.Point.fromHex(compressedViewingPublicKey.slice(2)).toHex()}`;
    return { spendingPublicKey, viewingPublicKey };
}
exports.getPublicKeys = getPublicKeys;
//# sourceMappingURL=cns.js.map