"use strict";
/**
 * @dev Functions for interacting with the Ethereum Name Service (ENS)
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.getPublicKeys = exports.namehash = exports.getResolverContract = void 0;
const ethers_1 = require("../ethers");
const KeyPair_1 = require("../classes/KeyPair");
const constants_1 = require("./constants");
const ENSRegistryAddress = '0x00000000000C2E074eC69A0dFb2997BA6C7d2e1e'; // same on all networks
/**
 * @notice Returns the address of the ENS stealth key resolver to use based on the provider's network
 * @param name ENS domain, e.g. myname.eth
 * @param provider Ethers provider
 */
const getResolverContract = async (name, provider) => {
    const registry = new ethers_1.Contract(ENSRegistryAddress, constants_1.ENS_REGISTRY_ABI, provider);
    const namehash = (0, ethers_1.namehash)(name);
    const resolverAddress = (await registry.resolver(namehash));
    // When using this method we expect the user will have an Umbra-compatible StealthKey resolver. There are two types:
    //   1. ForwardingStealthKeyResolver
    //   2. PublicStealthKeyResolver
    // Both can be found in this repo: https://github.com/ScopeLift/ens-resolvers
    //
    // Regardless of which type the user has, the ABI for getting and setting stealth keys is the same. Therefore
    // it's ok that we use always use the same ABI here)
    return new ethers_1.Contract(resolverAddress, constants_1.FORWARDING_STEALTH_KEY_RESOLVER_ABI, provider);
};
exports.getResolverContract = getResolverContract;
/**
 * @notice Computes ENS namehash of the input ENS domain, normalized to ENS compatibility
 * @param name ENS domain, e.g. myname.eth
 */
function namehash(name) {
    if (typeof name !== 'string') {
        throw new Error('Name must be a string');
    }
    return (0, ethers_1.namehash)(name);
}
exports.namehash = namehash;
/**
 * @notice For a given ENS domain, returns the public keys, or throws if they don't exist
 * @param name ENS domain, e.g. myname.eth
 * @param provider Ethers provider
 */
async function getPublicKeys(name, provider) {
    // Get the resolver used by `name`. If it's the zero address, they have no stealth keys
    const resolver = await (0, exports.getResolverContract)(name, provider);
    if (resolver.address === ethers_1.AddressZero) {
        throw new Error(`Name ${name} is not registered or user has not set their resolver`);
    }
    // Attempt to read stealth keys from the resolver contract
    let keys;
    try {
        keys = (await resolver.stealthKeys(namehash(name)));
    }
    catch (err) {
        throw new Error(`The configured resolver for ${name} does not support stealth keys. Please ask them to setup their Umbra account`); // prettier-ignore
    }
    // Make sure the found keys are not zero
    if (keys.spendingPubKey.eq(ethers_1.Zero) || keys.viewingPubKey.eq(ethers_1.Zero)) {
        throw new Error(`Public keys not found for ${name}. Please ask them to setup their Umbra account`);
    }
    // Uncompress keys and return them
    const spendingPublicKey = KeyPair_1.KeyPair.getUncompressedFromX(keys.spendingPubKey, keys.spendingPubKeyPrefix.toNumber());
    const viewingPublicKey = KeyPair_1.KeyPair.getUncompressedFromX(keys.viewingPubKey, keys.viewingPubKeyPrefix.toNumber());
    return { spendingPublicKey, viewingPublicKey };
}
exports.getPublicKeys = getPublicKeys;
//# sourceMappingURL=ens.js.map